# """This is temporary file that should be moved to mlem.gto module"""
# from typing import Optional

# from git import Repo
# from mlem.core.errors import MlemObjectNotFound
# from mlem.core.metadata import load_meta
# from mlem.core.objects import DatasetMeta, MlemMeta, ModelMeta
# from pydantic import BaseModel

# from gto.ext import Enrichment, EnrichmentInfo


# class MlemInfo(EnrichmentInfo):
#     source = "mlem"
#     meta: MlemMeta

#     def get_object(self) -> BaseModel:
#         return self.meta

#     def get_human_readable(self) -> str:
#         # TODO: create `.describe` method in MlemMeta https://github.com/iterative/mlem/issues/98
#         description = f"""Mlem {self.meta.object_type}"""
#         if isinstance(self.meta, ModelMeta):
#             description += f": {self.meta.model_type.type}"
#         if isinstance(self.meta, DatasetMeta):
#             description += f": {self.meta.dataset.dataset_type.type}"
#         return description


# class MlemEnrichment(Enrichment):
#     source = "mlem"

#     def describe(self, repo, obj: str, rev: Optional[str]) -> Optional[MlemInfo]:
#         try:
#             if isinstance(repo, Repo):
#                 repo = repo.working_dir
#             return MlemInfo(meta=load_meta(obj, repo=repo))  # rev=rev
#         except MlemObjectNotFound:
#             return None



import os
from time import sleep
from typing import Callable, Tuple

import git
import pytest

import gto
from gto.config import CONFIG_FILE_NAME


@pytest.fixture
def empty_git_repo(tmp_path):
    repo = git.Repo.init(tmp_path)

    def write_file(name, content):
        fullpath = os.path.join(tmp_path, name)
        os.makedirs(os.path.dirname(fullpath), exist_ok=True)
        with open(fullpath, "w", encoding="utf8") as file:
            file.write(content)

    return repo, write_file


@pytest.fixture
def repo_with_commit(empty_git_repo: Tuple[git.Repo, Callable]):
    repo, write_file = empty_git_repo

    write_file(
        "some-random-file",
        "some-random-text",
    )
    repo.index.add(["some-random-file"])
    repo.index.commit("Initial commit")

    return repo, write_file


@pytest.fixture
def init_showcase_semver(empty_git_repo: Tuple[git.Repo, Callable]):
    repo, write_file = empty_git_repo

    write_file(
        CONFIG_FILE_NAME,
        "",
    )

    return repo, write_file


@pytest.fixture
def showcase(
    init_showcase_semver,
):  # pylint: disable=too-many-locals, too-many-statements
    repo: git.Repo
    repo, write_file = init_showcase_semver
    path = repo.working_dir

    write_file("models/random-forest.pkl", "1st version")
    write_file("models/neural-network.pkl", "1st version")
    repo.index.add(["models"])
    repo.index.commit("Create models")

    gto.api.annotate(
        path, "rf", type="model", path="models/random-forest.pkl", must_exist=True
    )
    gto.api.annotate(
        path, "nn", type="model", path="models/neural-network.pkl", must_exist=True
    )
    gto.api.annotate(path, "features", type="dataset", path="datasets/features.csv")

    repo.index.add(["artifacts.yaml"])
    first_commit = repo.index.commit("Add artifacts")

    nn_vname = "v0.0.1"
    rf_vname = "v1.2.3"
    gto.api.register(path, "rf", "HEAD", rf_vname)
    gto.api.register(path, "nn", "HEAD")
    sleep(1)

    write_file("models/random-forest.pkl", "2nd version")

    second_commit = repo.index.commit("Update model")

    # bump version automatically
    gto.api.register(path, "rf", "HEAD")

    gto.api.promote(path, "nn", "staging", promote_version=nn_vname)
    gto.api.promote(path, "rf", "production", promote_version=rf_vname)
    sleep(1)  # this is needed to ensure right order of promotions in later checks
    # the problem is git tags doesn't have miliseconds precision, so we need to wait a bit
    gto.api.promote(path, "rf", "staging", promote_ref="HEAD")
    sleep(1)
    gto.api.promote(path, "rf", "production", promote_ref=repo.head.ref.commit.hexsha)
    sleep(1)
    gto.api.promote(path, "rf", "production", promote_version=rf_vname, force=True)
    return path, repo, write_file, first_commit, second_commit



import os
from typing import Callable, Tuple

import git
import pytest
from typer.testing import CliRunner

from gto.api import annotate
from gto.cli import app
from gto.config import CONFIG_FILE_NAME
from gto.exceptions import UnknownType
from gto.index import init_index_manager
from gto.registry import GitRegistry


@pytest.fixture
def init_repo(empty_git_repo: Tuple[git.Repo, Callable]):
    repo, write_file = empty_git_repo

    write_file(
        CONFIG_FILE_NAME,
        "type_allowed: [model, dataset]",
    )
    return repo


def test_config_load_index(init_repo):
    index = init_index_manager(init_repo)
    assert index.config.TYPE_ALLOWED == ["model", "dataset"]


def test_config_load_registry(init_repo):
    registry = GitRegistry.from_repo(init_repo)
    assert registry.config.TYPE_ALLOWED == ["model", "dataset"]


def test_adding_allowed_type(init_repo):
    annotate(init_repo, "name", type="model")


def test_adding_not_allowed_type(init_repo):
    with pytest.raises(UnknownType):
        annotate(init_repo, "name", type="unknown")


def test_config_is_not_needed(empty_git_repo: Tuple[git.Repo, Callable], request):
    repo, write_file = empty_git_repo

    write_file(
        CONFIG_FILE_NAME,
        "WRONG_CONFIG",
    )
    os.chdir(repo.working_dir)
    runner = CliRunner()
    result = runner.invoke(app, ["--help"])
    os.chdir(request.config.invocation_dir)
    assert result.exit_code == 0



from functools import total_ordering

import semver

from gto.exceptions import IncomparableVersions, InvalidVersion, WrongArgs


class AbstractVersion:
    version: str

    def __init__(self, version) -> None:
        if not self.is_valid(version):
            raise InvalidVersion(f"Supplied version '{version}' cannot be parsed")
        self.version = version

    @classmethod
    def is_valid(cls, version) -> bool:
        raise NotImplementedError

    def __eq__(self, other):
        return self.version == other.version

    def __lt__(self, other):
        raise NotImplementedError

    @classmethod
    def get_minimal(cls):
        raise NotImplementedError


@total_ordering
class SemVer(AbstractVersion):
    """
    A subclass of Version which allows a "v" prefix
    """

    @classmethod
    def is_valid(cls, version):
        try:
            cls.parse(version)
            return True
        except (InvalidVersion, ValueError, IndexError) as _:
            return False

    @classmethod
    def parse(cls, version: str) -> "SemVer":
        """
        Parse version string to a Version instance.

        :param version: version string with "v" or "V" prefix
        :raises ValueError: when version does not start with "v" or "V"
        :return: a new instance
        """
        if not isinstance(version, str):
            raise InvalidVersion("Version should be of type str")
        if version[0] not in ("v"):
            raise InvalidVersion(
                f"{version}: not a valid semantic version tag. Must start with 'v'"
            )
        return semver.VersionInfo.parse(version[1:])

    def __eq__(self, other):
        if isinstance(other, str):
            other = self.__class__(other)
        if not isinstance(other, self.__class__):
            raise IncomparableVersions(self, other)
        return self.version == other.version

    def __lt__(self, other):
        if isinstance(other, str):
            other = self.__class__(other)
        if not isinstance(other, self.__class__):
            raise IncomparableVersions(self, other)
        return self.parse(self.version) < self.parse(other.version)

    def bump_major(self):
        return self.__class__(f"v{self.parse(self.version).bump_major()}")

    def bump_minor(self):
        return self.__class__(f"v{self.parse(self.version).bump_minor()}")

    def bump_patch(self):
        return self.__class__(f"v{self.parse(self.version).bump_patch()}")

    def bump(self, bump_major=False, bump_minor=False, bump_patch=False):
        if sum(bool(i) for i in (bump_major, bump_minor, bump_patch)) != 1:
            raise WrongArgs("Need to specify exactly one bump argument")
        if bump_major:
            return self.bump_major()
        if bump_minor:
            return self.bump_minor()
        if bump_patch:
            return self.bump_patch()
        # TODO: stop using WrongArgs everywhere :)
        raise WrongArgs(
            "At least one of bump_major, bump_minor, bump_patch must be True"
        )

    @classmethod
    def get_minimal(cls):
        return cls("v0.0.1")



"""TODO: add more tests for API"""
from typing import Callable, Tuple

import git
import pytest

import gto
from gto.exceptions import WrongArgs
from gto.versions import SemVer
from tests.utils import _check_obj


def test_empty_index(empty_git_repo: Tuple[git.Repo, Callable]):
    repo, write_file = empty_git_repo  # pylint: disable=unused-variable
    index = gto.api.get_index(repo.working_dir)
    assert len(index.artifact_centric_representation()) == 0


def test_empty_state(empty_git_repo: Tuple[git.Repo, Callable]):
    repo, write_file = empty_git_repo  # pylint: disable=unused-variable
    state = gto.api._get_state(repo.working_dir)  # pylint: disable=protected-access
    assert len(state.artifacts) == 0


def test_api_info_commands_empty_repo(empty_git_repo: Tuple[git.Repo, Callable]):
    repo, write_file = empty_git_repo  # pylint: disable=unused-variable
    gto.api.show(repo.working_dir)
    gto.api.history(repo.working_dir)


def test_add_remove(empty_git_repo: Tuple[git.Repo, Callable]):
    repo, write_file = empty_git_repo  # pylint: disable=unused-variable
    name, type, path, must_exist = "new-artifact", "new-type", "new/path", False
    gto.api.annotate(
        repo.working_dir, name, type=type, path=path, must_exist=must_exist
    )
    index = gto.api.get_index(repo.working_dir).get_index()
    assert name in index
    _check_obj(
        index.state[name],
        dict(
            type=type,
            path=path,
            virtual=not must_exist,
            labels=[],
            description="",
        ),
        [],
    )
    gto.api.remove(repo.working_dir, name)
    index = gto.api.get_index(repo.working_dir).get_index()
    assert name not in index


@pytest.fixture
def repo_with_artifact(init_showcase_semver):
    repo: git.Repo
    repo, write_file = init_showcase_semver  # pylint: disable=unused-variable
    name, type, path, must_exist = "new-artifact", "new-type", "new/path", False
    gto.api.annotate(
        repo.working_dir, name, type=type, path=path, must_exist=must_exist
    )
    repo.index.add(["artifacts.yaml"])
    repo.index.commit("Added index")
    return repo, name


def test_register(repo_with_artifact):
    repo, name = repo_with_artifact
    vname1, vname2 = "v1.0.0", "v1.0.1"
    gto.api.register(repo.working_dir, name, "HEAD", vname1)
    latest = gto.api.find_latest_version(repo.working_dir, name)
    assert latest.name == vname1
    gto.api.annotate(
        repo.working_dir,
        "something-irrelevant",
        "doesnt-matter",
        "anything",
        must_exist=False,
    )
    repo.index.commit("Irrelevant action to create a git commit")
    gto.api.register(repo.working_dir, name, "HEAD")
    latest = gto.api.find_latest_version(repo.working_dir, name)
    assert latest.name == vname2


def test_promote(repo_with_artifact: Tuple[git.Repo, str]):
    repo, name = repo_with_artifact
    stage = "staging"
    repo.create_tag("v1.0.0")
    repo.create_tag("wrong-tag-unrelated")
    gto.api.promote(
        repo.working_dir, name, stage, promote_ref="HEAD", name_version="v0.0.1"
    )
    promotion = gto.api.find_promotion(repo.working_dir, name, stage)
    author = repo.commit().author.name
    author_email = repo.commit().author.email
    _check_obj(
        promotion,
        dict(
            artifact=name,
            version="v0.0.1",
            stage=stage,
            author=author,
            author_email=author_email,
            commit_hexsha=repo.commit().hexsha,
        ),
        {"created_at", "promotions", "tag"},
    )


def test_promote_skip_registration(repo_with_artifact):
    repo, name = repo_with_artifact
    stage = "staging"
    with pytest.raises(WrongArgs):
        gto.api.promote(
            repo.working_dir,
            name,
            stage,
            promote_ref="HEAD",
            name_version="v0.0.1",
            skip_registration=True,
        )
    gto.api.promote(
        repo.working_dir, name, stage, promote_ref="HEAD", skip_registration=True
    )
    promotion = gto.api.find_promotion(repo.working_dir, name, stage)
    assert not SemVer.is_valid(promotion.version)



import pytest

from gto.versions import SemVer


@pytest.mark.parametrize(
    "version", ["v0.1.2", "v3.4.5-pre.2+build.4", "v3.4.5-pre.2-build.4"]
)
def test_semver_is_valid(version):
    assert SemVer.is_valid(version)


@pytest.mark.parametrize(
    "version", [1, "1", "v1", "v1.2", "0.1.2", "3.4.5-pre.2+build.4"]
)
def test_semver_is_not_valid(version):
    assert not SemVer.is_valid(version)


def test_semver_comparison():
    assert SemVer("v1.3.2") < SemVer("v1.3.3")
    assert SemVer("v1.2.4") == SemVer("v1.2.4")
    assert SemVer("v1.3.4") < "v2.0.0"
    assert SemVer("v1.3.4") == "v1.3.4"


def test_bump_semver():
    assert SemVer("v1.3.4").bump_major() == SemVer("v2.0.0")
    assert SemVer("v1.3.4").bump_minor() == SemVer("v1.4.0")
    assert SemVer("v1.3.4").bump_patch() == SemVer("v1.3.5")



"""
Loggers used in other parts of GTO
"""
import logging.config

from gto.config import CONFIG

LOG_LEVEL = CONFIG.LOG_LEVEL
if CONFIG.DEBUG:
    LOG_LEVEL = logging.getLevelName(logging.DEBUG)

logging_config = {
    "version": 1,
    "disable_existing_loggers": False,
    "formatters": {
        "standard": {"format": "%(asctime)s [%(levelname)s] %(name)s: %(message)s"},
    },
    "handlers": {
        "default": {
            "level": LOG_LEVEL,
            "formatter": "standard",
            "class": "logging.StreamHandler",
            "stream": "ext://sys.stdout",
        },
    },
    "loggers": {
        "gto": {
            "handlers": ["default"],
            "level": LOG_LEVEL,
        }
    },
}

logging.config.dictConfig(logging_config)



import logging
from datetime import datetime
from enum import Enum
from typing import FrozenSet, Iterable, Optional, Union

import git
from pydantic import BaseModel

from gto.config import assert_name_is_valid, check_name_is_valid
from gto.versions import SemVer

from .base import (
    BaseArtifact,
    BaseManager,
    BasePromotion,
    BaseRegistryState,
    BaseVersion,
)
from .constants import ACTION, NAME, NUMBER, STAGE, TAG, VERSION, Action
from .exceptions import (
    InvalidTagName,
    InvalidVersion,
    MissingArg,
    RefNotFound,
    TagExists,
    UnknownAction,
)

ActionSign = {
    Action.REGISTER: "@",
    Action.PROMOTE: "#",
}


def name_tag(
    action: Action,
    name: str,
    version: Optional[str] = None,
    stage: Optional[str] = None,
    repo: Optional[git.Repo] = None,
    simple: bool = False,
):
    if action == Action.REGISTER:
        return f"{name}{ActionSign[action]}{version}"

    if action == Action.PROMOTE:
        if simple:
            return f"{name}{ActionSign[action]}{stage}"
        if repo is None:
            raise MissingArg(arg="repo")
        numbers = []
        for tag in repo.tags:
            parsed = parse_name(tag.name, raise_on_fail=False)
            if (
                parsed
                and (parsed[NAME] == name)
                and (parsed[ACTION] == Action.PROMOTE)
                and (NUMBER in parsed)
            ):
                numbers.append(parsed[NUMBER])
        new_number = max(numbers) + 1 if numbers else 1
        return f"{name}{ActionSign[action]}{stage}{ActionSign[action]}{new_number}"
    raise UnknownAction(action=action)


def _parse_register(name: str, raise_on_fail: bool = True):
    sign = len(name) - name[::-1].index(ActionSign[Action.REGISTER])
    name, version = name[: sign - 1], name[sign:]
    if check_name_is_valid(name) and SemVer.is_valid(version):
        return {
            ACTION: Action.REGISTER,
            NAME: name,
            VERSION: version,
        }
    if raise_on_fail:
        assert_name_is_valid(name)
        if not SemVer.is_valid(version):
            raise InvalidVersion(
                f"Version format should be v1.0.0, cannot parse {version}"
            )
    return {}


def _parse_promote(name: str, raise_on_fail: bool = True):
    parsed = name.split(ActionSign[Action.PROMOTE])
    if (
        check_name_is_valid(parsed[0])
        and check_name_is_valid(parsed[1])
        and (parsed[2].isdigit() if len(parsed) == 3 else True)
    ):
        if len(parsed) == 2:
            return {
                ACTION: Action.PROMOTE,
                NAME: parsed[0],
                STAGE: parsed[1],
            }
        if len(parsed) == 3:
            return {
                ACTION: Action.PROMOTE,
                NAME: parsed[0],
                STAGE: parsed[1],
                NUMBER: int(parsed[2]),
            }
    if raise_on_fail:
        assert_name_is_valid(parsed[0])
        assert_name_is_valid(parsed[1])
        if not parsed[2].isdigit():
            raise InvalidTagName(name)
    return {}


def parse_name(name: str, raise_on_fail: bool = True):

    if ActionSign[Action.REGISTER] in name:
        return _parse_register(name, raise_on_fail=raise_on_fail)

    if ActionSign[Action.PROMOTE] in name:
        return _parse_promote(name, raise_on_fail=raise_on_fail)

    if raise_on_fail:
        raise InvalidTagName(name)
    return {}


class NAME_REFERENCE(Enum):
    TAG = TAG
    NAME = NAME


def parse_name_reference(name: str):
    parsed = parse_name(name, raise_on_fail=False)
    if not parsed:
        return NAME_REFERENCE.NAME, name
    return NAME_REFERENCE.TAG, parsed


class Tag(BaseModel):
    action: Action
    name: str
    version: Optional[str]
    stage: Optional[str]
    created_at: datetime
    tag: git.Tag

    class Config:
        arbitrary_types_allowed = True


def parse_tag(tag: git.Tag):
    return Tag(
        tag=tag,
        created_at=datetime.fromtimestamp(tag.tag.tagged_date),
        **parse_name(tag.name),
    )


def find(
    action: Union[Action, FrozenSet[Action]] = None,
    name: Optional[str] = None,
    version: Optional[str] = None,
    stage: Optional[str] = None,
    repo: Optional[git.Repo] = None,
    sort: str = "by_time",
    tags: Optional[Iterable[git.Tag]] = None,
):
    if isinstance(action, Action):
        action = frozenset([action])
    if tags is None:
        if repo is None:
            raise MissingArg(arg="repo")
        tags = [t for t in repo.tags if parse_name(t.name, raise_on_fail=False)]
    if action:
        tags = [t for t in tags if parse_name(t.name)[ACTION] in action]
    if name:
        tags = [t for t in tags if parse_name(t.name).get(NAME) == name]
    if version:
        tags = [t for t in tags if parse_name(t.name).get(VERSION) == version]
    if stage:
        tags = [t for t in tags if parse_name(t.name).get(STAGE) == stage]
    if sort == "by_time":
        tags = sorted(tags, key=lambda t: t.tag.tagged_date)
    else:
        raise NotImplementedError(f"Unknown sort: {sort}")
    return tags


def create_tag(repo, name, ref, message):
    if all(c.hexsha != ref for c in repo.iter_commits()):
        raise RefNotFound(ref=ref)
    if name in repo.refs:
        raise TagExists(name=name)
    repo.create_tag(
        name,
        ref=ref,
        message=message,
    )


def version_from_tag(tag: git.Tag) -> BaseVersion:
    mtag = parse_tag(tag)
    return BaseVersion(
        artifact=mtag.name,
        name=mtag.version,
        created_at=mtag.created_at,
        author=tag.tag.tagger.name,
        author_email=tag.tag.tagger.email,
        commit_hexsha=tag.commit.hexsha,
        tag=tag.name,
    )


def promotion_from_tag(
    artifact: BaseArtifact, tag: git.Tag, version_required: bool
) -> BasePromotion:
    mtag = parse_tag(tag)
    if version_required:
        version = artifact.find_version(
            commit_hexsha=tag.commit.hexsha, raise_if_not_found=True
        ).name  # type: ignore
    else:
        version = artifact.find_version(commit_hexsha=tag.commit.hexsha)
        if version:
            version = version.name  # type: ignore
        else:
            artifact.add_version(
                BaseVersion(
                    artifact=mtag.name,
                    name=tag.commit.hexsha,
                    created_at=mtag.created_at,
                    author=tag.tag.tagger.name,
                    author_email=tag.tag.tagger.email,
                    commit_hexsha=tag.commit.hexsha,
                )
            )
            version = tag.commit.hexsha
    return BasePromotion(
        artifact=mtag.name,
        version=version,
        stage=mtag.stage,
        created_at=mtag.created_at,
        author=tag.tag.tagger.name,
        author_email=tag.tag.tagger.email,
        commit_hexsha=tag.commit.hexsha,
        tag=tag.name,
    )


def index_tag(
    artifact: BaseArtifact, tag: git.Tag, version_required: bool
) -> BaseArtifact:
    mtag = parse_tag(tag)
    if mtag.action == Action.REGISTER:
        artifact.add_version(version_from_tag(tag))
    if mtag.action == Action.PROMOTE:
        artifact.add_promotion(promotion_from_tag(artifact, tag, version_required))
    return artifact


class TagManager(BaseManager):  # pylint: disable=abstract-method
    def update_state(self, state: BaseRegistryState) -> BaseRegistryState:
        # tags are sorted and then indexed by timestamp
        # this is important to check that history is not broken
        tags = [parse_tag(t) for t in find(repo=self.repo, action=self.actions)]
        for tag in tags:
            state.update_artifact(
                index_tag(
                    state.find_artifact(tag.name, create_new=True),
                    tag.tag,
                    version_required=False,
                )
            )
        return state


class TagVersionManager(TagManager):
    actions: FrozenSet[Action] = frozenset((Action.REGISTER,))

    def register(self, name, version, ref, message):
        create_tag(
            self.repo,
            name_tag(Action.REGISTER, name, version=version, repo=self.repo),
            ref=ref,
            message=message,
        )

    def check_ref(self, ref: str, state: BaseRegistryState):
        try:
            _ = self.repo.tags[ref]
            art_name = parse_name(ref)[NAME]
            version_name = parse_name(ref)[VERSION]
        except (KeyError, ValueError, IndexError):
            logging.warning(
                "Provided ref doesn't exist or it is not a tag that registers a version"
            )
            return {}
        return {
            name: version
            for name, artifact in state.get_artifacts().items()
            for version in artifact.versions
            if name == art_name and version.name == version_name
        }


class TagStageManager(TagManager):
    actions: FrozenSet[Action] = frozenset((Action.PROMOTE,))

    def promote(self, name, stage, ref, message, simple):
        create_tag(
            self.repo,
            name_tag(Action.PROMOTE, name, stage=stage, repo=self.repo, simple=simple),
            ref=ref,
            message=message,
        )

    def check_ref(self, ref: str, state: BaseRegistryState):
        try:
            tag = self.repo.tags[ref]
            _ = parse_name(ref)[STAGE]
            art_name = parse_name(ref)[NAME]
        except (KeyError, ValueError, IndexError):
            logging.warning(
                "Provided ref doesn't exist or it is not a tag that promotes to an stage"
            )
            return {}
        return {
            name: promotion
            for name, artifact in state.get_artifacts().items()
            for promotion in artifact.stages
            if name == art_name
            and promotion.commit_hexsha == tag.commit.hexsha
            and promotion.created_at == datetime.fromtimestamp(tag.tag.tagged_date)
        }



from datetime import datetime
from typing import Dict, FrozenSet, List, Optional, Union

import git
from pydantic import BaseModel

from gto.config import RegistryConfig
from gto.constants import Action
from gto.ext import Enrichment
from gto.versions import SemVer

from .exceptions import ArtifactNotFound, ManyVersions, VersionRequired


class BasePromotion(BaseModel):
    artifact: str
    version: str
    stage: str
    created_at: datetime
    author: str
    author_email: Optional[str]  # remove as optional later
    commit_hexsha: str
    tag: str


class BaseVersion(BaseModel):
    artifact: str
    name: str
    created_at: datetime
    author: str
    author_email: Optional[str]  # remove as optional later
    commit_hexsha: str
    discovered: bool = False
    tag: Optional[str] = None
    promotions: List[BasePromotion] = []
    enrichments: List[Enrichment] = []

    @property
    def is_registered(self):
        """Tells if this is an explicitly registered version"""
        return self.tag is not None  # SemVer.is_valid(self.name)

    @property
    def version(self):
        # TODO: this should be read from config, how to pass it down here?
        return SemVer(self.name)

    @property
    def stage(self):
        promotions = sorted(self.promotions, key=lambda p: p.created_at)
        return promotions[-1] if promotions else None

    def dict_status(self):
        version = self.dict(exclude={"promotions"})
        version["stage"] = self.stage.dict() if self.stage else None
        return version


class BaseArtifact(BaseModel):
    name: str
    versions: List[BaseVersion]

    @property
    def stages(self):
        return [l for v in self.versions for l in v.promotions]

    @property
    def unique_stages(self):
        return {l.stage for l in self.stages}

    def __repr__(self) -> str:
        versions = ", ".join(f"'{v.name}'" for v in self.versions)
        stages = ", ".join(f"'{l}'" for l in self.unique_stages)
        return f"Artifact(versions=[{versions}], stages=[{stages}])"

    def get_latest_version(self, registered=False) -> Optional[BaseVersion]:
        versions = sorted(
            (
                v
                for v in self.versions
                if not v.discovered and (not registered or v.is_registered)
            ),
            key=lambda x: x.created_at,
        )
        if versions:
            return versions[-1]
        return None

    def get_promotions(self) -> Dict[str, BasePromotion]:
        stages: Dict[str, BasePromotion] = {}
        # semver
        versions = sorted(
            (v for v in self.versions if v.is_registered),
            key=lambda x: x.version,
            reverse=True,
        )
        # TODO: regular commits - these should be sorted by stage tag creation maybe?
        # E.g. when you created "rf#prod" tag, then you created a version.
        # It's probably should be a flag on CLI that will tell how to sort promoted for return.
        versions.extend(
            sorted(
                (v for v in self.versions if not v.is_registered),
                key=lambda x: x.created_at,
                reverse=True,
            )
        )
        for version in versions:
            promotion = version.stage
            if promotion:
                stages[promotion.stage] = stages.get(promotion.stage) or promotion
        return stages

    def add_version(self, version: BaseVersion):
        self.versions.append(version)

    def add_promotion(self, promotion: BasePromotion):
        self.find_version(name=promotion.version).promotions.append(  # type: ignore
            promotion
        )

    def update_enrichments(self, version, enrichments):
        self.find_version(
            name=version, include_discovered=True
        ).enrichments = enrichments

    @property
    def discovered(self):
        return any(not v.discovered for v in self.versions)

    # @overload
    # def find_version(
    #     self,
    #     name: str = None,
    #     commit_hexsha: str = None,
    #     raise_if_not_found: Literal[True] = ...,
    #     allow_multiple: Literal[False] = ...,
    # ) -> BaseVersion:
    #     ...

    # @overload
    # def find_version(
    #     self,
    #     name: str = None,
    #     commit_hexsha: str = None,
    #     raise_if_not_found: Literal[False] = ...,
    #     allow_multiple: Literal[False] = ...,
    # ) -> Optional[BaseVersion]:
    #     ...

    # @overload
    # def find_version(
    #     self,
    #     name: str = None,
    #     commit_hexsha: str = None,
    #     raise_if_not_found: Literal[False] = ...,
    #     allow_multiple: Literal[True] = ...,
    # ) -> List[BaseVersion]:
    #     ...

    def get_versions(self, include_non_explicit=False, include_discovered=False):
        return [
            v
            for v in self.versions
            if (v.is_registered and not v.discovered)
            or (include_discovered and v.discovered)
            or (include_non_explicit and not v.is_registered)
        ]

    def find_version(
        self,
        name: str = None,
        commit_hexsha: str = None,
        raise_if_not_found: bool = False,
        allow_multiple=False,
        include_discovered=False,
    ) -> Union[None, BaseVersion, List[BaseVersion]]:
        versions = [
            v
            for v in self.versions
            if (v.name == name if name else True)
            and (v.commit_hexsha == commit_hexsha if commit_hexsha else True)
            and (True if include_discovered else not v.discovered)
        ]

        if allow_multiple:
            return versions
        if raise_if_not_found and not versions:
            raise VersionRequired(name=self.name)
        if len(versions) > 1:
            raise ManyVersions(
                name=self.name,
                versions=[v.name for v in versions],
            )
        return versions[0] if versions else None

    def find_version_at_commit(
        self, commit_hexsha: str, latest_datetime: datetime = None
    ):
        return [
            v
            for v in self.find_version(  # type: ignore
                commit_hexsha=commit_hexsha,
                raise_if_not_found=True,
                allow_multiple=True,
            )
            if (v.created_at <= latest_datetime if latest_datetime else True)  # type: ignore
        ][-1]


class BaseRegistryState(BaseModel):
    artifacts: Dict[str, BaseArtifact] = {}

    class Config:
        arbitrary_types_allowed = True

    def add_artifact(self, name):
        self.artifacts[name] = BaseArtifact(name=name, versions=[])

    def update_artifact(self, artifact: BaseArtifact):
        self.artifacts[artifact.name] = artifact

    def get_artifacts(self):
        return self.artifacts

    def find_artifact(self, name: str, create_new=False):
        if not name:
            raise ValueError("Artifact name is required")
        if name not in self.artifacts:
            if create_new:
                self.artifacts[name] = BaseArtifact(name=name, versions=[])
            else:
                raise ArtifactNotFound(name)
        return self.artifacts.get(name)

    @property
    def unique_stages(self):
        return sorted({l for o in self.artifacts.values() for l in o.unique_stages})

    def find_commit(self, name, version):
        return (
            self.find_artifact(name)
            .find_version(name=version, raise_if_not_found=True)
            .commit_hexsha
        )

    def which(self, name, stage, raise_if_not_found=True):
        """Return stage active in specific stage"""
        promoted = self.find_artifact(name).get_promotions()
        if stage in promoted:
            return promoted[stage]
        if raise_if_not_found:
            raise ValueError(f"Stage {stage} not found for {name}")
        return None

    def sort(self):
        for name in self.artifacts:  # pylint: disable=consider-using-dict-items
            self.artifacts[name].versions.sort(key=lambda x: (x.created_at, x.name))
            self.artifacts[name].stages.sort(
                key=lambda x: (x.created_at, x.version, x.stage)
            )


class BaseManager(BaseModel):
    repo: git.Repo
    actions: FrozenSet[Action]
    config: RegistryConfig

    class Config:
        arbitrary_types_allowed = True

    def update_state(
        self, state: BaseRegistryState
    ) -> BaseRegistryState:  # pylint: disable=no-self-use
        raise NotImplementedError

    # better to uncomment this for typing, but it breaks the pylint

    # def register(self, name, version, ref, message):
    #     raise NotImplementedError

    # def promote(self, name, stage, ref, message):
    #     raise NotImplementedError

    def check_ref(self, ref: str, state: BaseRegistryState):
        raise NotImplementedError
